@@ -69,13 +69,9 @@ struct _DownloaderViewPrivate
 	GtkWidget *abort_button;
 
 	GtkStatusIcon *status_icon;
-
-#ifdef HAVE_LIBNOTIFY
-	NotifyNotification *notification;
-#endif
+	gboolean ownref;
 
 	guint source_id;
-	guint notification_timeout;
 };
 
 enum
@@ -120,10 +116,17 @@ download_dialog_delete_event_cb (GtkWidget *window,
 				 GdkEventAny *event,
 				 DownloaderView *dv);
 
-#ifdef HAVE_LIBNOTIFY
 static void
-show_notification_window (DownloaderView *dv);
-#endif
+download_progress_cb (WebKitDownload *download,
+		      GParamSpec *pspec,
+		      DownloaderView *dv);
+static void
+download_status_changed_cb (WebKitDownload *download,
+			    GParamSpec *pspec,
+			    DownloaderView *dv);
+static gboolean
+download_error_cb (WebKitDownload *download, gint error_code, gint error_detail,
+		   const gchar *reason, DownloaderView *dv);
 
 G_DEFINE_TYPE (DownloaderView, downloader_view, EPHY_TYPE_DIALOG)
 
@@ -181,32 +184,36 @@ show_status_icon (DownloaderView *dv)
 
 	priv->status_icon = gtk_status_icon_new_from_stock (STOCK_DOWNLOAD);
 
-#ifdef HAVE_LIBNOTIFY
-	notify_notification_attach_to_status_icon (priv->notification, priv->status_icon);
-#endif
-
 	g_signal_connect_swapped (priv->status_icon, "activate",
 				  G_CALLBACK (show_downloader_cb), dv);
 	g_signal_connect (priv->status_icon, "popup-menu",
 			  G_CALLBACK (status_icon_popup_menu_cb), dv);
+
+	gtk_status_icon_set_visible (priv->status_icon, TRUE);
 }
 
 static gboolean
 remove_download (WebKitDownload *download,
 		 gpointer rowref,
-		 DownloaderView *view)
+		 DownloaderView *dv)
 {
 	WebKitDownloadStatus status;
 
-	g_signal_handlers_disconnect_matched
-		(download, G_SIGNAL_MATCH_DATA ,
-		 0, 0, NULL, NULL, view);
+	g_signal_handlers_disconnect_by_func
+		(download, G_CALLBACK (download_progress_cb), dv);
+
+	g_signal_handlers_disconnect_by_func
+		(download, G_CALLBACK (download_status_changed_cb), dv);
+
+	g_signal_handlers_disconnect_by_func
+		(download, G_CALLBACK (download_error_cb), dv);
 
 	status = webkit_download_get_status (download);
 	if (status == WEBKIT_DOWNLOAD_STATUS_STARTED)
 		webkit_download_cancel (download);
 
 	g_object_unref (download);
+	g_object_unref (dv);
 	return TRUE;
 }
 
@@ -243,6 +250,8 @@ downloader_view_init (DownloaderView *dv)
 
 	downloader_view_build_ui (dv);
 
+	dv->priv->ownref = TRUE;
+
 	show_status_icon (dv);
 
 	g_signal_connect_object (embed_shell, "prepare_close",
@@ -272,11 +281,6 @@ downloader_view_finalize (GObject *object)
 		priv->source_id = 0;
 	}
 	
-	if (priv->notification_timeout != 0)
-	{
-		g_source_remove (priv->notification_timeout);
-	}
-
 	g_hash_table_destroy (dv->priv->downloads_hash);
 
 	G_OBJECT_CLASS (downloader_view_parent_class)->finalize (object);
@@ -292,6 +296,65 @@ downloader_view_new (void)
 			 			   NULL));
 }
 
+#ifdef HAVE_LIBNOTIFY
+static void
+notification_closed_cb (NotifyNotification *notification,
+			DownloaderView *dv)
+{
+	g_object_unref (dv);
+}
+
+static gboolean
+show_notification_cb (NotifyNotification *notification)
+{
+	DownloaderView *dv;
+	GError *error = NULL;
+
+	dv = g_object_get_data (G_OBJECT (notification), "dv");
+	notify_notification_show (notification, &error);
+
+	if (error)
+	{
+		/* notification_closed_cb () will never be called. */
+		g_warning ("Error showing download notification: %s",
+			   error->message);
+		g_object_unref (dv);
+		g_error_free (error);
+	}
+
+	return FALSE;
+}
+
+static void
+show_notification (DownloaderView *dv, const char *title, const char *msg)
+{
+	NotifyNotification *notification;
+	GtkStatusIcon *status_icon;
+
+	status_icon = dv->priv->status_icon;
+
+	/* We keep the DownloaderView alive until the notification is gone. */
+	g_object_ref (dv);
+
+	notification = notify_notification_new (title, msg,
+						GTK_STOCK_INFO, NULL);
+
+	g_signal_connect_after (notification, "closed",
+				G_CALLBACK (notification_closed_cb), dv);
+	g_object_set_data (G_OBJECT (notification), "dv", dv);
+
+	notify_notification_set_timeout (notification, NOTIFY_EXPIRES_DEFAULT);
+	notify_notification_set_urgency (notification, NOTIFY_URGENCY_LOW);
+
+	notify_notification_attach_to_status_icon (notification, status_icon);
+
+	/* There are some visual glitches when the notification is shown and
+	 * the GtkStatusIcon is still not visible. To avoid that, we delay the
+	 * popup a bit. */
+	g_timeout_add (500, (GSourceFunc) show_notification_cb, notification);
+}
+#endif
+
 static char *
 format_interval (gdouble interval)
 {
@@ -366,13 +429,17 @@ static char *
 ephy_download_get_name (WebKitDownload *download)
 {
 	const char *target;
+	char *user_destination;
 	char *result;
 
 	target = webkit_download_get_destination_uri (download);
+	user_destination = g_object_get_data (G_OBJECT (download),
+					      "user-destination-uri");
 
-	if (target)
+	if (user_destination || target)
 	{
-		result = g_path_get_basename (target);
+		result = g_path_get_basename (user_destination ?
+					      user_destination : target);
 	}
 	else
 	{
@@ -410,13 +477,22 @@ do_open_downloaded_file (DownloaderView *dv, WebKitDownload *download, gboolean
 {
 	DownloaderViewPrivate *priv = dv->priv;
 	GdkDisplay *gdk_display;
-	const char *destination_uri = webkit_download_get_destination_uri (download);
-	GFile *downloaded_file = g_file_new_for_uri (destination_uri);
+	const char *destination_uri;
+	char *user_destination;
+	GFile *downloaded_file;
 
 	gdk_display = gtk_widget_get_display (priv->window);
+
+	destination_uri = webkit_download_get_destination_uri (download);
+	user_destination = g_object_get_data (G_OBJECT (download),
+					      "user-destination-uri");
+
+	downloaded_file = g_file_new_for_uri (user_destination ?
+					      user_destination : destination_uri);
 	if (open_location)
 	{
-		ephy_file_browse_to (downloaded_file, gdk_x11_display_get_user_time (gdk_display));
+		ephy_file_browse_to (downloaded_file,
+				     gdk_x11_display_get_user_time (gdk_display));
 	}
 	else
 	{
@@ -445,7 +521,7 @@ update_download_row (DownloaderView *dv, WebKitDownload *download)
 	GtkTreePath *path;
 	GtkTreeIter iter;
 	WebKitDownloadStatus status;
-	gint64 total, current;
+	guint64 total, current;
 	gdouble remaining_seconds = 0.0;
 	char *remaining, *file, *cur_progress, *name;
 	struct tm;
@@ -456,7 +532,12 @@ update_download_row (DownloaderView *dv, WebKitDownload *download)
 #endif
 
 	row_ref = get_row_from_download (dv, download);
-	g_return_if_fail (row_ref != NULL);
+	/* In downloader_view_add_download() we call this function manually to
+	 * ensure the view has something visible in the tree view, however
+	 * signal handlers might have already removed this download from the
+	 * DV, this is that case. */
+	if (row_ref == NULL)
+		return;
 
 	/* Status special casing */
 	status = webkit_download_get_status (download);
@@ -485,24 +566,19 @@ update_download_row (DownloaderView *dv, WebKitDownload *download)
 		case DOWNLOAD_ACTION_OPEN_LOCATION:
 			open_downloaded_file_location (dv, download);
 			break;
+		case DOWNLOAD_ACTION_DOWNLOAD:
+			break;
 		default:
 			g_assert_not_reached ();
 			break;
 		}
 
-		downloader_view_remove_download (dv, download);
 #ifdef HAVE_LIBNOTIFY
-		downloaded = g_strdup_printf (_("The file “%s” has been downloaded."), 
-						name);
-		notify_notification_update (dv->priv->notification,
-			_("Download finished"), 
-			downloaded, 
-			GTK_STOCK_INFO);
-			
-		show_notification_window (dv);
-		
+		downloaded = g_strdup_printf (_("The file “%s” has been downloaded."), name);
+		show_notification (dv, _("Download finished"), downloaded);
 		g_free (downloaded);
 #endif
+		downloader_view_remove_download (dv, download);
 
 		return;
 	case WEBKIT_DOWNLOAD_STATUS_STARTED:
@@ -591,10 +667,12 @@ download_status_changed_cb (WebKitDownload *download, GParamSpec *pspec, Downloa
                 update_download_row (dv, download);
 }
 
-static void
+static gboolean
 download_error_cb (WebKitDownload *download, gint error_code, gint error_detail, const gchar *reason, DownloaderView *dv)
 {
 	update_download_row (dv, download);
+
+	return FALSE;
 }
 
 static gboolean
@@ -606,34 +684,6 @@ update_buttons_timeout_cb (DownloaderView *dv)
 	return FALSE;
 }
 
-#ifdef HAVE_LIBNOTIFY
-static gboolean
-queue_show_notification (DownloaderView *dv)
-{
-	if (gtk_status_icon_is_embedded (dv->priv->status_icon))
-	{
-		notify_notification_show (dv->priv->notification, NULL);
-		dv->priv->notification_timeout = 0;
-		return FALSE;
-	}
-
-	return TRUE;
-}
-
-static void
-show_notification_window (DownloaderView *dv)
-{
-	if (gtk_status_icon_is_embedded (dv->priv->status_icon))
-		notify_notification_show (dv->priv->notification, NULL);
-	else
-	{
-		if (dv->priv->notification_timeout != 0)
-			g_source_remove (dv->priv->notification_timeout);
-		dv->priv->notification_timeout = g_timeout_add_seconds (1, (GSourceFunc) queue_show_notification, dv);
-	}
-}
-#endif
-
 void
 downloader_view_add_download (DownloaderView *dv,
 			      WebKitDownload *download)
@@ -649,15 +699,11 @@ downloader_view_add_download (DownloaderView *dv,
 	char *mime, *icon_name;
 	int width = 16, height = 16;
 #endif
-	GValue visible = {0, };
-
-#ifdef HAVE_LIBNOTIFY
-	char *downloading;
-#endif
+	gboolean visible = FALSE;
 
 	/* dv may be unrefed inside update_download_row if the file
 	 * downloaded completely while the user was choosing where to
-	 * put it, so we need to protect it
+	 * put it, so we need to protect it.
 	 */
 	g_object_ref (dv);
 	g_object_ref (download);
@@ -675,11 +721,10 @@ downloader_view_add_download (DownloaderView *dv,
 			     download,
 			     row_ref);
 
-	update_download_row (dv, download);
 	update_status_icon (dv);
 
 	selection = gtk_tree_view_get_selection
-		(GTK_TREE_VIEW(dv->priv->treeview));
+		(GTK_TREE_VIEW (dv->priv->treeview));
 	gtk_tree_selection_unselect_all (selection);
 	gtk_tree_selection_select_iter (selection, &iter);
 
@@ -692,25 +737,33 @@ downloader_view_add_download (DownloaderView *dv,
 	g_signal_connect_object (download, "error",
 				 G_CALLBACK (download_error_cb), dv, 0);
 
+	update_download_row (dv, download);
+
 	/* Show it already */
-	g_value_init (&visible, G_TYPE_BOOLEAN);
-	g_object_get_property (G_OBJECT (dv->priv->window), "visible", &visible);
+	g_object_get (G_OBJECT (dv->priv->window), "visible", &visible, NULL);
+
+	/* In the previous update_download_row() (or handlers) the download
+	 * might have finished, if that's the case, we have nothing else to do.
+	 * A notification of the finished download will pop to inform the user
+	 * there was success. */
+	if (g_hash_table_size (dv->priv->downloads_hash) < 1)
+	{
+		ephy_dialog_hide (EPHY_DIALOG (dv));
+		return;
+	}
 	
-	if (eel_gconf_get_boolean (CONF_DOWNLOADS_HIDDEN) && !g_value_get_boolean (&visible))
+	if (eel_gconf_get_boolean (CONF_DOWNLOADS_HIDDEN) && !visible)
 	{
-
 #ifdef HAVE_LIBNOTIFY
-		char *name = ephy_download_get_name (download);
-		downloading = g_strdup_printf(_("The file “%s” has been added to the downloads queue."), 
-						name);
-		g_free (name);
-		notify_notification_update (dv->priv->notification,
-					_("Download started"), 
-					downloading, 
-					GTK_STOCK_INFO);
-		
-		show_notification_window (dv);
+		char *name;
+		char *downloading;
 
+		name = ephy_download_get_name (download);
+		downloading = g_strdup_printf(_("The file “%s” has been added to the downloads queue."), name);
+		
+		show_notification (dv, _("Download started"), downloading);
+
+		g_free (name);
 		g_free (downloading);
 #endif
 
@@ -720,8 +773,6 @@ downloader_view_add_download (DownloaderView *dv,
 	{
 		ephy_dialog_show (EPHY_DIALOG (dv));
 	}
-	
-	g_value_unset (&visible);
 
 #if 0
         // FIXMEchpe port this to use GIcon when webkit gets download support
@@ -752,10 +803,8 @@ downloader_view_add_download (DownloaderView *dv,
 	}
 #endif
 
-	dv->priv->source_id = g_timeout_add (100, (GSourceFunc) update_buttons_timeout_cb, dv);
-
-	/* see above */
-	g_object_unref (dv);
+	if (dv->priv->source_id == 0)
+	    dv->priv->source_id = g_timeout_add (100, (GSourceFunc) update_buttons_timeout_cb, dv);
 }
 
 static void
@@ -904,13 +953,6 @@ downloader_view_build_ui (DownloaderView *dv)
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview));
 	gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);
 	g_signal_connect (selection, "changed", G_CALLBACK (selection_changed), dv);
-	
-#ifdef HAVE_LIBNOTIFY
-	priv->notification = notify_notification_new (" ", " ", GTK_STOCK_INFO, NULL);
-	notify_notification_set_timeout (priv->notification, NOTIFY_EXPIRES_DEFAULT);
-	notify_notification_set_urgency (priv->notification, NOTIFY_URGENCY_LOW);
-#endif
-
 }
 
 static void
@@ -997,10 +1039,17 @@ downloader_view_remove_download (DownloaderView *dv, WebKitDownload *download)
 	update_status_icon (dv);
 
 	/* Close the dialog if there are no more downloads */
-
 	if (!g_hash_table_size (dv->priv->downloads_hash))
 	{
-		g_object_unref (dv);
+		gtk_widget_set_sensitive (dv->priv->abort_button, FALSE);
+		gtk_widget_set_sensitive (dv->priv->pause_button, FALSE);
+		ephy_dialog_hide (EPHY_DIALOG (dv));
+
+		if (dv->priv->ownref)
+		{
+			dv->priv->ownref = FALSE;
+			g_object_unref (dv);
+		}
 	}
 }
 
